@@ -58,7 +58,7 @@ errcode_t e2fsck_reset_context(e2fsck_t ctx)
 		ext2fs_free_icount(ctx->inode_link_info);
 		ctx->inode_link_info = 0;
 	}
-	if (ctx->fs->dblist) {
+	if (ctx->fs && ctx->fs->dblist) {
 		ext2fs_free_dblist(ctx->fs->dblist);
 		ctx->fs->dblist = 0;
 	}
@@ -153,6 +153,8 @@ int e2fsck_run(e2fsck_t ctx)
 		if (ctx->flags & E2F_FLAG_RUN_RETURN)
 			break;
 		e2fsck_pass(ctx);
+		if (ctx->progress)
+			(void) (ctx->progress)(ctx, 0, 0, 0);
 	}
 	ctx->flags &= ~E2F_FLAG_SETJMP_OK;
 	
