@@ -675,6 +675,16 @@ static const struct e2fsck_problem problem_table[] = {
 	  "@d @e for '.' is big.  ",
 	  PROMPT_SPLIT, PR_NO_OK },
 
+	/* Illegal FIFO inode */
+	{ PR_2_BAD_FIFO,
+	  "@i %i (%Q) is an @I FIFO.\n",
+	  PROMPT_CLEAR, 0 },
+
+	/* Illegal socket inode */
+	{ PR_2_BAD_SOCKET,
+	  "@i %i (%Q) is an @I socket.\n",
+	  PROMPT_CLEAR, 0 },
+
 	/* Pass 3 errors */
 
 	/* Pass 3: Checking directory connectivity */
@@ -700,7 +710,7 @@ static const struct e2fsck_problem problem_table[] = {
 	/* /lost+found not found */
 	{ PR_3_NO_LF_DIR,
 	  "/@l not found.  ",
-	  PROMPT_CREATE, 0 },
+	  PROMPT_CREATE, PR_PREEN_OK },
 
 	/* .. entry is incorrect */
 	{ PR_3_BAD_DOT_DOT,
@@ -787,6 +797,11 @@ static const struct e2fsck_problem problem_table[] = {
 	  "Cannot proceed without a @r.\n",
 	  PROMPT_NONE, PR_FATAL },  
 
+	/* Internal error: couldn't find dir_info */
+	{ PR_3_NO_DIRINFO,
+	  "Internal error: couldn't find dir_info for %i.\n",
+	  PROMPT_NONE, PR_FATAL },
+
 	/* Pass 4 errors */
 	
 	/* Pass 4: Checking reference counts */
