@@ -402,6 +402,12 @@ struct problem_context {
 /* Directory entry for '.' is big.  Split? */
 #define PR_2_SPLIT_DOT		0x0200024
 
+/* Illegal FIFO */
+#define PR_2_BAD_FIFO		0x020025
+
+/* Illegal socket */
+#define PR_2_BAD_SOCKET		0x020026
+
 /*
  * Pass 3 errors
  */
@@ -472,6 +478,9 @@ struct problem_context {
 /* Cannot proceed without a root inode. */
 #define PR_3_NO_ROOT_INODE_ABORT	0x030015
 
+/* Internal error: couldn't find dir_info */
+#define PR_3_NO_DIRINFO			0x020016
+
 /*
  * Pass 4 errors
  */
