@@ -4,24 +4,18 @@
  * For copyright info, see mit-sipb-copyright.h.
  */
 
+#include "com_err.h"
 #include "error_table.h"
 #include "mit-sipb-copyright.h"
 #include "internal.h"
 
-#ifndef	lint
-static const char copyright[] =
-    "Copyright 1987,1988 by Student Information Processing Board, Massachusetts Institute of Technology";
-static const char rcsid_et_name_c[] =
-    "$Header: /mit/krb5/.cvsroot/src/util/et/et_name.c,v 5.0 1993/04/13 19:56:23 tytso Exp $";
-#endif
-
 static const char char_set[] =
 	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_";
 
 static char buf[6];
 
 const char * error_table_name(num)
-    int num;
+    errcode_t num;
 {
     int ch;
     int i;
@@ -31,10 +25,10 @@ const char * error_table_name(num)
     p = buf;
     num >>= ERRCODE_RANGE;
     /* num = ?? ??? ??? aaa aaa bbb bbb ccc ccc ddd ddd */
-    num &= 077777777;
+    num &= 077777777L;
     /* num = 00 000 000 aaa aaa bbb bbb ccc ccc ddd ddd */
     for (i = 4; i >= 0; i--) {
-	ch = (num >> BITS_PER_CHAR * i) & ((1 << BITS_PER_CHAR) - 1);
+	ch = (int)((num >> BITS_PER_CHAR * i) & ((1 << BITS_PER_CHAR) - 1));
 	if (ch != 0)
 	    *p++ = char_set[ch-1];
     }
