@@ -310,7 +310,7 @@ ext2_off_t ext2fs_file_get_size(ext2_file_t file)
  * 
  * XXX still need to call truncate
  */
-extern errcode_t ext2fs_file_set_size(ext2_file_t file, ext2_off_t size)
+errcode_t ext2fs_file_set_size(ext2_file_t file, ext2_off_t size)
 {
 	errcode_t	retval;
 	EXT2_CHECK_MAGIC(file, EXT2_ET_MAGIC_EXT2_FILE);
