@@ -42,7 +42,7 @@
 #endif
 
 #include "et/com_err.h"
-#include "ext2fs/io.h"
+#include "ext2fs/ext2_io.h"
 
 const char * program_name = "badblocks";
 
@@ -50,7 +50,7 @@ int v_flag = 0;			/* verbose */
 int w_flag = 0;			/* do r/w test */
 int s_flag = 0;			/* show progress of test */
 
-static volatile void usage (void)
+static void usage(void)
 {
 	fprintf (stderr, "Usage: %s [-b block_size] [-o output_file] [-svw] device blocks_count\n [start_count]\n",
 		 program_name);
@@ -260,9 +260,9 @@ static void test_rw (int dev, unsigned long blocks_count,
 	}
 }
 
-void main (int argc, char ** argv)
+int main (int argc, char ** argv)
 {
-	char c;
+	int c;
 	char * tmp;
 	char * device_name;
 	char * output_file = NULL;
@@ -298,14 +298,14 @@ void main (int argc, char ** argv)
 			w_flag = 1;
 			break;
 		default:
-			usage ();
+			usage();
 		}
 	}
 	if (optind > argc - 1)
-		usage ();
+		usage();
 	device_name = argv[optind++];
 	if (optind > argc - 1)
-		usage ();
+		usage();
 	blocks_count = strtoul (argv[optind], &tmp, 0);
 	if (*tmp)
 	{
@@ -346,4 +346,5 @@ void main (int argc, char ** argv)
 	close (dev);
 	if (out != stdout)
 		fclose (out);
+	exit(0);
 }
