@@ -41,7 +41,7 @@
 const char * program_name = "dumpe2fs";
 char * device_name = NULL;
 
-static volatile void usage (void)
+static void usage(void)
 {
 	fprintf (stderr, "usage: %s [-bV] device\n", program_name);
 	exit (1);
@@ -172,13 +172,13 @@ static int i386_byteorder(void)
 	return (*cp == 1);
 }
 
-void main (int argc, char ** argv)
+int main (int argc, char ** argv)
 {
 	errcode_t	retval;
 	ext2_filsys	fs;
 	int		print_badblocks = 0;
 	int		big_endian;
-	char		c;
+	int		c;
 
 	initialize_ext2_error_table();
 	fprintf (stderr, "dumpe2fs %s, %s for EXT2 FS %s, %s\n",
@@ -198,11 +198,11 @@ void main (int argc, char ** argv)
 				error_message(EXT2_ET_BASE));
 			exit(0);
 		default:
-			usage ();
+			usage();
 		}
 	}
 	if (optind > argc - 1)
-		usage ();
+		usage();
 	device_name = argv[optind++];
 	retval = ext2fs_open (device_name, 0, 0, 0, unix_io_manager, &fs);
 	if (retval) {
