@@ -36,11 +36,9 @@
 #ifdef HAVE_MNTENT_H
 #include <mntent.h>
 #endif
-#include <malloc.h>
 #include <sys/ioctl.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <stdio.h>
 
 #ifdef HAVE_LINUX_FS_H
 #include <linux/fs.h>
@@ -55,6 +53,9 @@
 #include "ext2fs/ext2fs.h"
 #include "../version.h"
 
+/* Everything is STDC, these days */
+#define NOARGS void
+
 #define STRIDE_LENGTH 8
 
 #ifndef sparc
@@ -81,14 +82,16 @@ char *creator_os = NULL;
 char *volume_label = NULL;
 char *mount_dir = NULL;
 
+static void usage(NOARGS), check_plausibility(NOARGS), check_mount(NOARGS);
+
 static void usage(NOARGS)
 {
 	fprintf(stderr, "Usage: %s [-c|-t|-l filename] [-b block-size] "
 	"[-f fragment-size]\n\t[-i bytes-per-inode] "
-	"[-m reserved-blocks-percentage] [-qvS]\n\t"
-	"[-o creator-os] [-g blocks-per-group] [-L volume-label]\n\t"
-	"[-M last-mounted-directory] [-r fs-revision] [-R raid_opts]\n\t"
-	"device [blocks-count]\n",
+	" [-N number-of-inodes]\n\t[-m reserved-blocks-percentage] "
+	"[-o creator-os] [-g blocks-per-group]\n\t[-L volume-label] "
+	"[-M last-mounted-directory] [-r fs-revision]\n\t[-R raid_opts]"
+	"[-s sparse-super-flag] [-qvSV] device [blocks-count]\n",
 		program_name);
 	exit(1);
 }
@@ -105,6 +108,28 @@ static int log2(int arg)
 	return l;
 }
 
+static int log10(unsigned int arg)
+{
+	int	l;
+
+	for (l=0; arg ; l++)
+		arg = arg / 10;
+	return l;
+}
+
+static void proceed_question(NOARGS)
+{
+	fflush(stdout);
+	fflush(stderr);
+	printf("Proceed anyway? (y,n) ");
+	if (getchar() != 'y')
+		exit(1);
+}
+
+#ifndef SCSI_BLK_MAJOR
+#define SCSI_BLK_MAJOR(M)  ((M) == SCSI_DISK_MAJOR || (M) == SCSI_CDROM_MAJOR)
+#endif
+
 static void check_plausibility(NOARGS)
 {
 #ifdef HAVE_LINUX_MAJOR_H
@@ -114,29 +139,24 @@ static void check_plausibility(NOARGS)
 	val = stat(device_name, &s);
 	
 	if(val == -1) {
-		printf("Could not stat %s --- %s\n", device_name,
-		       error_message(errno));
+		fprintf(stderr, "Could not stat %s --- %s\n",
+			device_name, error_message(errno));
 		if (errno == ENOENT)
-			printf("\nThe device apparently does not exist; "
-			       "did yo specify it correctly?\n");
+			fprintf(stderr, "\nThe device apparently does "
+			       "not exist; did you specify it correctly?\n");
 		exit(1);
 	}
 	if(!S_ISBLK(s.st_mode)) {
 		printf("%s is not a block special device.\n", device_name);
-		printf("Proceed anyway? (y,n) ");
-		if (getchar() != 'y')
-			exit(1);
+		proceed_question();
 		return;
-	}
-	if ((MAJOR(s.st_rdev) == HD_MAJOR && MINOR(s.st_rdev)%64 == 0) ||
-	    (MAJOR(s.st_rdev) == SCSI_DISK_MAJOR &&
-	     MINOR(s.st_rdev)%16 == 0)) {
+	} else if ((MAJOR(s.st_rdev) == HD_MAJOR &&
+		    MINOR(s.st_rdev)%64 == 0) ||
+		   (SCSI_BLK_MAJOR(MAJOR(s.st_rdev)) &&
+		       MINOR(s.st_rdev)%16 == 0)) {
 		printf("%s is entire device, not just one partition!\n", 
 		       device_name);
-		printf("Proceed anyway? (y,n) ");
-		if (getchar() != 'y')
-			exit(1);
-		return;
+		proceed_question();
 	}
 #endif
 }
@@ -203,8 +223,8 @@ static void test_disk(ext2_filsys fs, badblocks_list *bb_list)
 	errcode_t	retval;
 	char		buf[1024];
 
-	sprintf(buf, "badblocks %s%s %d", quiet ? "" : "-s ",
-		fs->device_name,
+	sprintf(buf, "badblocks -b %d %s%s %d", fs->blocksize,
+		quiet ? "" : "-s ", fs->device_name,
 		fs->super->s_blocks_count);
 	if (verbose)
 		printf("Running command: %s\n", buf);
@@ -261,9 +281,9 @@ static void handle_bad_blocks(ext2_filsys fs, badblocks_list bb_list)
 	 */
 	group_block = fs->super->s_first_data_block +
 		fs->super->s_blocks_per_group;
-	group_bad = 0;
 	
 	for (i = 1; i < fs->group_desc_count; i++) {
+		group_bad = 0;
 		for (j=0; j < fs->desc_blocks+1; j++) {
 			if (badblocks_list_test(bb_list, group_block +
 						j)) {
@@ -301,6 +321,7 @@ static void write_inode_tables(ext2_filsys fs)
 	blk_t		blk;
 	int		i, j, num, count;
 	char		*buf;
+	char		format[20], backup[80];
 
 	buf = malloc(fs->blocksize * STRIDE_LENGTH);
 	if (!buf) {
@@ -308,12 +329,22 @@ static void write_inode_tables(ext2_filsys fs)
 		exit(1);
 	}
 	memset(buf, 0, fs->blocksize * STRIDE_LENGTH);
-	
+
+	/*
+	 * Figure out how many digits we need
+	 */
+	i = log10(fs->group_desc_count);
+	sprintf(format, "%%%dd/%%%dld", i, i);
+	memset(backup, '\b', sizeof(backup)-1);
+	backup[sizeof(backup)-1] = 0;
+	if ((2*i)+1 < sizeof(backup))
+		backup[(2*i)+1] = 0;
+
 	if (!quiet)
 		printf("Writing inode tables: ");
 	for (i = 0; i < fs->group_desc_count; i++) {
 		if (!quiet)
-			printf("%4d/%4ld", i, fs->group_desc_count);
+			printf(format, i, fs->group_desc_count);
 		
 		blk = fs->group_desc[i].bg_inode_table;
 		num = fs->inode_blocks_per_group;
@@ -330,11 +361,11 @@ static void write_inode_tables(ext2_filsys fs)
 				       count, blk, error_message(retval));
 		}
 		if (!quiet) 
-			printf("\b\b\b\b\b\b\b\b\b");
+			fputs(backup, stdout);
 	}
 	free(buf);
 	if (!quiet)
-		printf("done     \n");
+		fputs("done                            \n", stdout);
 }
 
 static void create_root_dir(ext2_filsys fs)
@@ -446,7 +477,7 @@ static void show_stats(ext2_filsys fs)
 	struct ext2fs_sb 	*s = (struct ext2fs_sb *) fs->super;
 	char 			buf[80];
 	blk_t			group_block;
-	int			i, col_left;
+	int			i, need, col_left;
 	
 	if (param.s_blocks_count != s->s_blocks_count)
 		printf("warning: %d blocks unused.\n\n",
@@ -490,10 +521,12 @@ static void show_stats(ext2_filsys fs)
 		group_block += s->s_blocks_per_group;
 		if (!ext2fs_bg_has_super(fs, i))
 			continue;
-		if (!col_left--) {
+		need = log10(group_block) + 2;
+		if (need > col_left) {
 			printf("\n\t");
-			col_left = 8;
+			col_left = 72;
 		}
+		col_left -= need;
 		printf("%u", group_block);
 		if (i != fs->group_desc_count - 1)
 			printf(", ");
@@ -538,7 +571,7 @@ static int set_os(struct ext2_super_block *sb, char *os)
 
 #define PATH_SET "PATH=/sbin"
 
-static parse_raid_opts(const char *opts)
+static void parse_raid_opts(const char *opts)
 {
 	char	*buf, *token, *next, *p, *arg;
 	int	len;
@@ -593,17 +626,19 @@ static parse_raid_opts(const char *opts)
 
 static void PRS(int argc, char *argv[])
 {
-	char	c;
+	int	c;
 	int	size;
 	char	* tmp;
 	blk_t	max = 8192;
 	int	inode_ratio = 4096;
 	int	reserved_ratio = 5;
+	ino_t	num_inodes = 0;
 	errcode_t	retval;
 	int	sparse_option = -1;
 	char	*oldpath = getenv("PATH");
 	struct ext2fs_sb *param_ext2 = (struct ext2fs_sb *) &param;
 	char	*raid_opts = 0;
+	blk_t	dev_size;
 	
 	/* Update our PATH to include /sbin  */
 	if (oldpath) {
@@ -628,7 +663,7 @@ static void PRS(int argc, char *argv[])
 	if (argc && *argv)
 		program_name = *argv;
 	while ((c = getopt (argc, argv,
-			    "b:cf:g:i:l:m:o:qr:R:s:tvI:SFL:M:")) != EOF)
+			    "b:cf:g:i:l:m:o:qr:R:s:tvI:SFL:M:N:V")) != EOF)
 		switch (c) {
 		case 'b':
 			size = strtoul(optarg, &tmp, 0);
@@ -711,6 +746,9 @@ static void PRS(int argc, char *argv[])
 			param.s_inode_size = atoi(optarg);
 			break;
 #endif
+		case 'N':
+			num_inodes = atoi(optarg);
+			break;
 		case 'v':
 			verbose = 1;
 			break;
@@ -732,6 +770,11 @@ static void PRS(int argc, char *argv[])
 		case 'S':
 			super_only = 1;
 			break;
+		case 'V':
+			/* Print version number and exit */
+			fprintf(stderr, "\tUsing %s\n",
+				error_message(EXT2_ET_BASE));
+			exit(0);
 		default:
 			usage();
 		}
@@ -759,15 +802,27 @@ static void PRS(int argc, char *argv[])
 
 	param.s_log_frag_size = param.s_log_block_size;
 
+	retval = ext2fs_get_device_size(device_name,
+					EXT2_BLOCK_SIZE(&param),
+					&dev_size);
+	if (retval && (retval != EXT2_ET_UNIMPLEMENTED)) {
+		com_err(program_name, retval,
+			"while trying to determine filesystem size");
+		exit(1);
+	}
 	if (!param.s_blocks_count) {
-		retval = ext2fs_get_device_size(device_name,
-						EXT2_BLOCK_SIZE(&param),
-						&param.s_blocks_count);
-		if (retval) {
-			com_err(program_name, retval,
-				"while trying to determine filesystem size");
+		if (retval == EXT2_ET_UNIMPLEMENTED) {
+			com_err(program_name, 0,
+				"Couldn't determine device size; you "
+				"must specify\nthe size of the "
+				"filesystem\n");
 			exit(1);
-		}
+		} else
+			param.s_blocks_count = dev_size;
+	} else if (!force && (param.s_blocks_count > dev_size)) {
+		com_err(program_name, 0,
+			"Filesystem larger than apparent filesystem size.");
+		proceed_question();
 	}
 
 	if (param.s_blocks_per_group) {
@@ -782,7 +837,7 @@ static void PRS(int argc, char *argv[])
 	/*
 	 * Calculate number of inodes based on the inode ratio
 	 */
-	param.s_inodes_count =
+	param.s_inodes_count = num_inodes ? num_inodes : 
 		((long long) param.s_blocks_count * EXT2_BLOCK_SIZE(&param))
 			/ inode_ratio;
 
@@ -798,7 +853,7 @@ static void PRS(int argc, char *argv[])
 #ifdef EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER
 	if ((sparse_option == 1)
 #ifdef EXT2_DYNAMIC_REV
-	    || (param.s_rev_level >= EXT2_DYNAMIC_REV) && (!sparse_option)
+	    || ((param.s_rev_level >= EXT2_DYNAMIC_REV) && (!sparse_option))
 #endif
 	    ) 
 		param_ext2->s_feature_ro_compat |=
